# -*- coding: utf-8 -*-
#
# MONK automated test framework
#
# Copyright (C) 2013 DResearch Fahrzeugelektronik GmbH
# Written and maintained by MONK Developers <project-monk@dresearch-fe.de>
#
# This program is free software; you can redistribute it and/or # modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version
# 3 of the License, or (at your option) any later version.
#

from nose import tools as nt
from monk_tf import conn


def test_simplest():
    """ conn: create the simplest possible AConnection
    """
    # nothing to prepare
    # execute
    sut = conn.AConnection()
    # assert
    nt.ok_(sut, "should contain an AConnection object, but contains '{}'"
            .format(sut))

def test_call_methods():
    """ conn: calling a public AConnection method calls its state's method
    """
    # prepare
    state = MockState()
    expected_calls = ["connect", "login", "cmd", "disconnect"]
    sut = conn.AConnection(start_state=state)
    # execute
    sut.connect()
    sut.login()
    sut.cmd(None)
    sut.disconnect()
    # assert
    nt.eq_(expected_calls, state.calls,
            "didn't call the following methods: \"{}\"".format(
                list(set(expected_calls) - set(state.calls))))

def test_fsm():
    """ conn: go through all state transitions
    """
    # prepare
    txt_in = "qwerty12345"
    expected = txt_in
    sut = conn.EchoConnection()
    sut.credentials = ("not", "important")
    # execute
    sut.connect()
    sut.login()
    after_login = sut.current_state
    out = sut.cmd(txt_in)
    sut.disconnect()
    # here no exceptions is already a good sign that fsm works
    # assert
    nt.eq_(sut.current_state, conn.Disconnected(),
            "after complete transition end state should be disconnected")
    nt.eq_(after_login, conn.Authenticated(),
            "after login, state should be authenticated")
    nt.eq_(out, expected, "cmd should return same message as was put in")

@nt.raises(conn.NotConnectedException)
def test_wrong_state():
    """ conn: raise Exception when sending cmd unconnected
    """
    # prepare
    sut = conn.EchoConnection()
    # execute
    sut.cmd("")
    # finished, because cmd should raise exception

def test_cmd_returncode():
    """ conn: test connections can handle additional parameters
    """
    # set up
    sut = conn.EchoConnection()
    sut2 = conn.DefectiveConnection()
    # execute + assert (raises Error if params can't be parsed)
    sut._cmd("hello", returncode=True)
    try:
        out = sut2._cmd("hello", returncode=True)
    except conn.MockConnectionException as e:
        pass

def test_connected_login():
    """ conn: connection's _login is not called if already logged in
    """
    # set up
    sut = MockConnection(start_state=conn.Authenticated())
    # execute
    sut.login()
    sut.login()
    sut.login()
    # assert
    nt.ok_("_login" not in sut.calls)

@nt.raises(conn.CantConnectException)
def test_legal_port():
    """ conn: using a non existing port results in exception
    """
    # setup
    sut = conn.SerialConnection(port="this/port/can/hopefully/not/exis.t")
    # exercise
    sut.connect()

@nt.raises(conn.CantConnectException)
def test_noprompt_exception():
    """ conn: connecting to shut down target device results in exception
    """
    # setup
    sut = conn.SilentConnection()
    # exercise
    sut.connect()

def test_noprompt_notconnected():
    """ conn: connecting to shut down target device doesn't change state
    """
    # setup
    sut = conn.SilentConnection()
    # exercise
    try:
        sut.connect()
    except conn.CantConnectException as e:
        # verify
        nt.ok_(isinstance(sut.current_state, conn.Disconnected))

class MockConnection(conn.AConnection):

    def __init__(self, *args, **kwargs):
        self.calls = []
        self.logged_in = kwargs.pop("logged_in", False)
        super(MockConnection, self).__init__(*args, **kwargs)

    def _connect(self):
        self.calls.add("_connect")

    def _login(self):
        self.calls.append("_login")
        self.logged_in = True

    def _cmd(self, *args, **kwargs):
        self.calls.add("_cmd")

    def _disconnect(self):
        self.calls.add("_disconnect")

class MockState(conn.AState):
    calls = []

    def connect(self, connection):
        self.calls.append("connect")

    def login(self, connection):
        self.calls.append("login")

    def cmd(self, connection, msg):
        self.calls.append("cmd")

    def disconnect(self, connection):
        self.calls.append("disconnect")

    def next_state(self, connection):
        return self



# -*- coding: utf-8 -*-
#
# MONK automated test framework
#
# Copyright (C) 2013 DResearch Fahrzeugelektronik GmbH
# Written and maintained by MONK Developers <project-monk@dresearch-fe.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version
# 3 of the License, or (at your option) any later version.
#

from nose import tools as nt
from monk_tf import dev
from monk_tf import conn

def test_simple():
    """ dev: create Device
    """
    # execute
    sut = dev.Device()
    # assert
    nt.ok_(sut, "should have an object of dev.Device")

def test_send_cmd():
    """ dev: send cmd to EchoConnection
    """
    # prepare
    test_input = "just some text"
    expected_out = test_input
    sut = dev.Device(conn.EchoConnection())
    # execute
    out = sut.cmd(test_input)
    # assert
    nt.eq_(expected_out, out)

def test_second_conn():
    """ dev: let the first connection fail
    """
    # prepare
    test_input = "another text with defective connection"
    expected_out = test_input
    sut = dev.Device(conn.DefectiveConnection(), conn.EchoConnection())
    # execute
    out = sut.cmd(test_input)
    # assert
    nt.eq_(expected_out, out)

@nt.raises(dev.CantHandleException)
def test_no_conn():
    """ dev: let all connections fail
    """
    # prepare
    test_input = "no connection will succeed here"
    expected_out = test_input
    sut = dev.Device(conn.DefectiveConnection(), conn.DefectiveConnection())
    # execute
    out = sut.cmd(test_input)
    # catch exception



# -*- coding: utf-8 -*-
#
# MONK automated test framework
#
# Copyright (C) 2013 DResearch Fahrzeugelektronik GmbH
# Written and maintained by MONK Developers <project-monk@dresearch-fe.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version # 3 of the License, or (at your option) any later version.  #

"""
This module implements connection handling. Using the classes from this module
you can connect directly to a :term:`target device` via serial or ssh.
Example::

    import monk_tf.conn as mc
    # create a serial connection
    serial=mc.SerialConn(port="/dev/ttyUSB3", user="tester", pw="test")
    # create a ssh connection
    ssh=mc.SshConn(host="192.168.2.123", user="tester", pw="test")
    # send a command
    print serial.cmd("ls -al")
    [...]
    # send a command
    ssh.cmd("ls -al")
    [...]
"""

import os
import sys
import re
import logging

import pexpect
from pexpect import fdpexpect

class ConnectionBase(object):
    """ is the base class for all connections.

    Don't instantiate this class directly.

    This class implements the behaviour of cmd() interactions, and it makes
    sure that it doesn't login a user that is already logged in.

    Extending this class requires to implement _get_exp() and _login().
    """



    def __init__(self):
        if hasattr(self, "name"):
            self._logger = logging.getLogger(self.name)
        else:
            self._logger = logging.getLogger(type(self).__name__)
        self._logger.debug("hi.")

    @property
    def exp(self):
        """ the pexpect object - Don't bother with this if you don't know what
                                 it means.
        """
        try:
            return self._exp
        except AttributeError as e:
            self._exp = self._get_exp()
            return self._exp

    def _expect(self, pattern, timeout=-1, searchwindowsize=-1):
        """ a wrapper for :pexpect:meth:`spawn.expect`
        """
        self._logger.debug("expect({},{},{})".format(
            str(pattern).encode('string-escape'), timeout, searchwindowsize))
        try:
            self.exp.expect(pattern, timeout, searchwindowsize)
            self._logger.debug("expect succeeded.")
        except Exception as e:
            self._logger.debug("expect failed.")
            raise e

    def _send(self, s):
        """ a wrapper for :pexpect:meth:`spawn.send`
        """
        self._logger.debug("send({})".format(s))
        try:
            self.exp.send(s)
            self._logger.debug("send succeeded.")
        except Exception as e:
            self._logger.debug("send failed.")
            raise e

    def _sendline(self, s=""):
        """ a wrapper for :pexpect:meth:`spawn.sendline`
        """
        self._logger.debug("sendline({})".format(s))
        try:
            self.exp.sendline(s)
            self._logger.debug("sendline succeeded.")
        except Exception as e:
            self._logger.debug("sendline failed.")
            raise e

    def login(self, user=None, pw=None, timeout=30):
        """ attempts to authenticate to the connection.

        Default for user and password are the one's given to the connection on
        instantiation.

        :param user: the username
        :param pw: the password
        :param timeout: how long the connection waits to see whether it is
                        logged in already
        """
        self._logger.debug("login({},{},{})".format(user, pw, timeout))
        try:
            self._sendline("")
            self._expect(self.prompt, timeout=timeout)
            self._logger.debug("already logged in")
        except pexpect.TIMEOUT as e:
            self._login(user, pw)

    def cmd(self, msg, expect=None, timeout=30, login_timeout=None):
        """ send a shell command and retreive its output.

        :param msg: the shell command
        :param expect: a regex that represents the end of an interaction.
                       Defaults to the prompt set on connection instantiation
        :param timeout: how long a command call should wait for its desired
                        result
        :param login_timeout: how long the connection should wait until it
                              decides a login is necessary.

        :return: the stdout and stderr of the shell command
        """
        self._logger.debug("cmd({},{},{},{})".format(
            msg, expect, timeout, login_timeout))
        self.login(timeout=login_timeout or timeout)
        self._sendline(msg)
        expect_msg = re.escape(msg[:5]) + "[^\n]*\r\n"
        self._expect(expect_msg, timeout=timeout)
        self._expect(expect or self.prompt, timeout=timeout)
        self._logger.debug("cmd({}) result='{}' expect-match='{}'".format(
            str(msg[:15]).encode("string_escape") + ("[...]" if len(msg) > 15 else ""),
            str(self.exp.before[:50]).encode("string-escape") + ("[...]" if len(self.exp.before) > 50 else ""),
            str(self.exp.after[:50]).encode("string-escape") + ("[...]" if len(self.exp.after) > 50 else ""),
        ))
        return self.exp.before

    def __del__(self):
        self._logger.debug("bye.")
        self.exp.close()

class SerialConn(ConnectionBase):
    """ implements a serial connection.
    """

    def __init__(self, name, port, user, pw, prompt="\r?\n?[^\n]*#"):
        """
        :param name: the name of the connection
        :param port: the path to the device file that is used for this connection
        :param user: the user name for the login
        :param pw: the password for the login
        :param prompt: the default prompt to check for
        """
        self.name = name
        self.port = port
        self.user = user
        self.pw = pw
        self.prompt = prompt
        super(SerialConn, self).__init__()

    def _get_exp(self):
        spawn = fdpexpect.fdspawn(os.open(self.port, os.O_RDWR|os.O_NONBLOCK|os.O_NOCTTY))
        #spawn.logfile = sys.stdout
        return spawn

    def _login(self, user=None, pw=None):
        self._logger.debug("serial._login({},{})".format(user, pw))
        self._expect("[lL]ogin: ")
        self._sendline(user or self.user)
        self._expect("[pP]assword: ")
        self._sendline(pw or self.pw)
        self._expect(self.prompt)

class SshConn(ConnectionBase):
    """ implements an ssh connection.
    """

    def __init__(self, name, host, user, pw, prompt="\r?\n?[^\n]*#"):
        """
        :param name: the name of the connection
        :param host: the URL to the device
        :param user: the user name for the login
        :param pw: the password for the login
        :param prompt: the default prompt to check for
        """
        self.name = name
        self.host= host
        self.user = user
        self.pw = pw
        self.prompt = prompt
        super(SshConn, self).__init__()

    def _get_exp(self):
        return pexpect.spawn("ssh {}@{} -o TCPKeepAlive=yes -o ServerAliveInterval=5 -o ServerAliveCountMax=3".format(
            self.user,
            self.host
        ))
    def _login(self, user=None, pw=None):
        self._logger.debug("ssh._login({},{})".format(user, pw))
        self._expect("[pP]assword: ")
        self._sendline(pw or self.pw)
        self._expect(self.prompt)



# -*- coding: utf-8 -*-
#
# MONK automated test framework
#
# Copyright (C) 2013 DResearch Fahrzeugelektronik GmbH
# Written and maintained by MONK Developers <project-monk@dresearch-fe.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version
# 3 of the License, or (at your option) any later version.
#

"""
Instead of creating :py:class:`~monk_tf.dev.Device` and
:py:class:`~monk_tf.conn.AConnection` objects by yourself, you can also choose
to put corresponding data in a separate file and let this layer handle the
object concstruction and destruction for you. Doing this will probably make
your test code look more clean, keep the number of places where you need to
change something as small as possible, and lets you reuse data that you already
have described.

A hello world test with it looks like this::

    import nose
    from monk_tf import fixture

    def test_hello():
        ''' say hello
        '''
        # set up
        h = fixture.Fixture('target_device.cfg')
        expected_out = "hello"
        # execute
        out = h.devs[0].cmd('echo "hello"')
        # assert
        nose.tools.eq_(expected_out, out)
        # tear down
        h.tear_down()

When using this layer setting up a device only takes one line of code. The rest
of the information is in the ``target_device.cfg`` file. :term:`MONK` currently 
comes with one text format parser predefined, which is the
:py:class:`~monk_tf.fixture.XiniParser`. ``Xini`` is short for
:term:`extended INI`. You may, however, use any data format you want, if you
extend the :py:class:`~monk_tf.fixture.AParser` class accordingly.

An example ``Xini`` data file might look like this::

    [device1]
        type=Device
        [[serial1]]
            type=SerialConnection
            port=/dev/ttyUSB1
            user=example
            password=secret

As you can see it looks like an :term:`INI` file. There are sections,
consisting of a title enclosed in squared brackets (``[]``) and lists of
properties, consisting of key-value pairs separated by equality signs (``=``).
The unusual part is that the section *serial1* is surrounded by two pairs of
squared brackets (``[]``). This is the specialty of this format indicating that
*serial1* is a subsection of *device1* and therefore is a nested section. This
nesting can be done unlimited, by surrounding a section with more and more
pairs of squared brackets (``[]``) according to the level of nesting intended.
In this example *serial1* belongs to *device1* and the types indicate the
corresponding :term:`MONK` object to be created.

Classes
-------
"""

import os
import sys
import logging

import configobj as config

import conn
import dev

logger = logging.getLogger(__name__)

############
#
# Exceptions
#
############

class AFixtureException(Exception):
    """ Base class for exceptions of the fixture layer.

    If you want to make sure that you catch all exceptions that are related
    to this layer, you should catch *AFixtureExceptions*. This also means
    that if you extend this list of exceptions you should inherit from this
    exception and not from :py:exc:`~exceptions.Exception`.
    """
    pass

class AParseException(AFixtureException):
    """ Base class for exceptions concerning parsing errors.
    """
    pass

class NotXiniException(AParseException):
    """
    is raised when a :term:`fixture file` could not be parsed as
    :term:`extended INI`.
    """
    pass

class CantParseException(AFixtureException):
    """ is raised when a Fixture cannot parse a given file.
    """
    pass

class NoDeviceException(AFixtureException):
    """ is raised when a :py:clas:`~monk_tf.fixture.Fixture` requires a device but has none.
    """

######################################################
#
# Parsers - Read a text file and be used like a dict()
#
######################################################

class AParser(dict):
    """ Base class for all parsers.

    Do not instantiate this class! This basically just provides the :term:`API`
    that is needed by :py:class:`~monk_tf.fixture.Fixture` to interact with the
    data that is parsed. Each child class should make sure that it always
    provides its parsed data like a :py:class:`dict` would. If you require your
    own parser, you can extend this. :py:class:`~monk_tf.fixture.XiniParser`
    provides a very basic example.
    """
    pass

class XiniParser(config.ConfigObj, AParser):
    """ Reads config files in :term:`extended INI` format.
    """

    def _load(self, infile, configspec):
        """ Changes exception type raised.

        Overwrites method from :py:class:`~configobj.ConfigObj` to raise a
        :py:class:`~monk_tf.fixture.NotXiniException` instead of a
        :py:class:`~configobj.ConfigObjError`.
        """
        try:
            self.file_error = True
            super(XiniParser, self)._load(infile, configspec)
        except config.ConfigObjError as e:
            t, val, traceback = sys.exc_info()
            raise NotXiniException, e.message, traceback

##############################################################
#
# Fixture Classes - creates MONK objects based on dictionaries
#
##############################################################

class Fixture(object):
    """ Creates :term:`MONK` objects based on dictionary like objects.

    This is the class that provides the fundamental feature of this layer. It
    reads data files by trying to parse them via its list of known parsers and
    if it succeeds, it creates :term:`MONK` objects based on the configuration
    given by the data file. Most likely these objects are one or more
    :py:class:`~monk_tf.dev.Device` objects that have at least one
    :py:class:`~monk_tf.conn.AConnection` object each. If more than one
    :term:`fixture file` is read containing the same name on the highest level,
    then the latest data gets used. This does not work on lower levels of
    nesting, though. If you attempt to overwrite lower levels of nesting, what
    actually happens is that the highest layer gets overwritten and you lose
    the data that was stored in the older objects. This is simply how
    :py:meth:`set.update` works.

    One source of data (either a file name or a child class of
    :py:class:`~monk_tf.fixture.AParser`) can be given to an object of this
    class by its constructer, others can be added afterwards with the
    :py:meth:`~monk_tf.fixture.Fixture.read` method. An example looks like
    this::

        import monk_tf.fixture as mf

        fixture = mf.Fixture('/etc/monk_tf/default_devices.cfg')
                .read('~/.monk/default_devices.cfg')
                # can also be a parser object
                .read(XiniParser('~/testsuite12345/suite_devices.cfg'))

    """

    _DEFAULT_CLASSES = {
        "Device" : dev.Device,
        "HydraDevice" : dev.Hydra,
        "SerialConnection" : conn.SerialConn,
        "SshConnection" : conn.SshConn,
    }

    _DEFAULT_PARSERS = [
        XiniParser,
    ]

    _DEFAULT_DEBUG_SOURCE = "MONK_DEBUG_SOURCE"

    def __init__(self, source=None, name=None, parsers=None, classes=None,
            lookfordbgsrc=True):
        """

        :param source: The :term:`fixture file` or
                       :py:class:`~monk_tf.fixture.AParser` object to be read.

        :param name: The name of this object.

        :param parsers: An :python:term:`iterable` of
                        :py:class:`~monk_tf.fixture.AParser` classes to be used
                        for parsing a given
                        :py:attr:`~monk_tf.fixture.Fixture.source`.

        :param classes: A :py:class:`dict` of classes to class names. Used for
                        parsing the type attribute in
                        :term:`fixture files<fixture file>`.

        :param lookfordbgsrc: If True an environment variable is looked for to
                              read a local debug config. If False it won't be
                              looked for.
        """
        self.name = name or self.__class__.__name__
        self._logger = logging.getLogger("{}:{}".format(__name__, self.name))
        self.devs = []
        self.parsers = parsers or self._DEFAULT_PARSERS
        self.classes = classes or self._DEFAULT_CLASSES
        self.props = {}
        if source:
            self._update_props(
                    self._parse(source))
        if lookfordbgsrc and self._DEFAULT_DEBUG_SOURCE in os.environ:
            self._logger.debug("load debug source from {}".format(
                self._DEFAULT_DEBUG_SOURCE))
            self._update_props(
                    self._parse(os.environ[self._DEFAULT_DEBUG_SOURCE]))
        else:
            self._logger.debug("no debug source file found")
        self._initialize()


    def read(self, source):
        """ Read more data, either as a file name or as a parser.

        :param source: the data source; either a file name or a
                       :py:class:`~monk_tf.fixture.AParser` child class
                       instance.

        :return: self
        """
        self._logger.debug("read: " + str(source))
        props = self._parse(source)
        self._update_props(props)
        self._initialize()
        return self

    def _parse(self, source):
        """ Parse data file.

        :param source: the data source; either a file name or a
                       :py:class:`~monk_tf.fixture.AParser` child class
                       instance.

        :return: Returns a :py:class:`~monk_tf.fixture.AParser` instance.
        :raises: :py:class:`~monk_tf.fixture.CantParseException`
        """
        self._logger.debug("parse: " + str(source))
        if isinstance(source, AParser):
            return source
        else:
            for parser in self.parsers:
                try:
                    return parser(source)
                except AParseException as e:
                    self._logger.exception(e)
                    continue
            raise CantParseException()

    def _update_props(self, props):
        """ Updates the properties with a dictionary-like object.

        This basically uses :py:meth:`dict.update` to update
        :py:attr:`self.props <~monk_tf.fixture.Fixture.props>` with a new set
        of data.

        :param props: object that can be used with :py:meth:`dict.update`
        """
        self._logger.debug("add props: " + str(props))
        self.props.update(props)
        self._logger.debug("final props: " + str(props))

    def _initialize(self):
        """ Create :term:`MONK` objects based on self's properties.
        """
        self._logger.debug("initialize with props: " + str(self.props))
        self.devs = []
        for dname in self.props.keys():
            dconf = dict(self.props[dname])
            dclass = self.classes[dconf.pop("type")]
            dconns = []
            for cname in dconf.keys():
                cconf = dict(dconf[cname])
                cclass = self.classes[cconf.pop("type")]
                cconf['name'] = cname
                dconns.append(cclass(**cconf))
            self.devs.append(dclass(name=dname, conns=dconns))

    def cmd(self, msg):
        """ call :py:meth:`cmd` from first :py:class:`~monk_tf.device.Device`
        """
        try:
            return self.devs[0].cmd(msg)
        except IndexError:
            raise NoDeviceException("this fixture has no device loaded")

    def tear_down(self):
        """ Can be used for explicit destruction of managed objects.

        This should be called in every :term:`test case` as the last step.

        """
        for device in self.devs:
            try:
                del device
            except Exception as e:
                logger.exception(e)
        self.devs = []

    def __str__(self):
        return "{cls}.devs:{devs}".format(
                cls=self.__class__.__name__,
                devs=[str(d) for d in self.devs],
        )
